/////////////////////////////////////////////////////////////////////
///   devnetX Quad Art-Net Node (DMX Output Firmware), Ver. 1.0   ///
///   DMX512 Transmitter Module                                   ///
///   ---------------------------------------------------------   ///
///   Author: Alexej Goujine                                      ///
///   Copyright  2011 - 2013 devnetX, http://www.devnetx.at      ///
///   E-Mail: dev@devnetx.at                                      ///
///   Last changed on: 02.12.2013 (DD.MM.YYYY)                    ///
///   ---------------------------------------------------------   ///
///   This code is provided 'as is', use it at you own risk.      ///
/////////////////////////////////////////////////////////////////////

#include <p18cxxx.h>
#include <GenericTypeDefs.h>
#include <string.h>
#include "DMX512.h"
#include "SPIDMA.h"

// Definitions
#define FOSC                48000000u
#define TMR0_PRESCALER      8u
#define DMXLED_BLINKSPEED   5u

#define DELAY_BREAK         256 - (0.000088 * FOSC) / (4 * TMR0_PRESCALER)     // 88s
#define DELAY_MAB           256 - (0.000008 * FOSC) / (4 * TMR0_PRESCALER)     // 8s
#define DELAY_FRAME         256 - (0.000044 * FOSC) / (4 * TMR0_PRESCALER)     // 44s

// Enums
typedef enum
{
    SEND_BREAK = 0,
    SEND_MAB = 1,
    SEND_STARTCODE = 2,
    SEND_DATA = 3
} DMXSTATEMACHINE;

// Variables
#pragma udata DMX
static volatile DMXPACKET DMXTXPacket;
#pragma udata
static volatile DMXPACKET *DMXTXPacketPointer;
static volatile DMXSTATEMACHINE DMXState;
static volatile UINT8 DMXLEDCounter;
static volatile BOOL DMXLEDBlinkerEnabled;

// Prototypes
static void DMX512EnableBlinker(void);
static void DMX512DisableBlinker(void);

// Initializes the DMX512 transmitter engine
void DMX512Init(void)
{
    // Initialize Pins
    DMX_LED = 0;
    DMX_LED_TRIS = OUTPUT_PIN;

    DMX_BREAK = 0;
    DMX_BREAK_TRIS = OUTPUT_PIN;

    // Initialize Pointers
    DMXTXPacketPointer = &DMXTXPacket;

    // Clear DMX Buffer
    memset(DMXTXPacketPointer, 0x00, sizeof(DMXTXPacket));

    // Set DMX Transfer State Machine
    DMXState = SEND_BREAK;
    DMXLEDCounter = 0;
    DMXLEDBlinkerEnabled = FALSE;

    // Initialize UART
    DMX_TXSTA = 0b01100101;    // BRGH = 1, TXEN = 1, TX9 = 1, TX9D = 1
    DMX_RCSTA = 0b10000000;    // SPEN = 1
    Nop(); Nop();              // Silicon Errata #3 workaround
    DMX_BAUDCON = 0b00001000;  // BRG16 = 1
    DMX_SPBRGH = 0;
    DMX_SPBRGL = 47;

    // Initialize Timer0
    TMR0L = 0;                  // Preset for Timer0
    T0CON = 0b11000010;         // Prescaler: 1:8, 8-Bit Mode, TMR0ON = 1

    // Initialize Timer1
    T1CON = 0b00110000;

    // Initialize Timer0 Interrupt
    INTCON2bits.TMR0IP = 1;     // Timer0 Interrupt Priority: HIGH
    INTCONbits.T0IF = 0;        // Clear Timer0 Interrupt Flag
    INTCONbits.T0IE = 1;        // Timer0 Interrupt: ENABLED

    // Initialize Timer1 Interrupt
    IPR1bits.TMR1IP = 0;        // Timer1 Interrupt Priority: LOW
    PIR1bits.TMR1IF = 0;        // Clear Timer1 Interrupt Flag
    PIE1bits.TMR1IE = 1;        // Timer1 Interrupt: ENABLED

    // Disable DMX LED Blinker
    DMX512DisableBlinker();
}

// Timer0 Interrupt for DMX transmitter engine
void DMX512Interrupt(void)
{
    static UINT16 DMXCurrentChannel = 0;

    // If there is no data to send, send the BREAK
    //if (DMXTXPacketPointer->DMXDataLength == 0)
    //    DMXState = SEND_BREAK;

    // Check the state of the DMX state machine
    switch (DMXState)
    {
        case SEND_DATA: // STARTCODE END - DATA START
            TMR0L = DELAY_FRAME;

            // Transmit current BYTE FRAME
            DMX_TXREG = DMXTXPacketPointer->DMXData[DMXCurrentChannel];
            DMXCurrentChannel++;

            // All BYTE FRAMES transmitted?
            if (DMXCurrentChannel >= DMXTXPacketPointer->DMXDataLength)
                DMXState = SEND_BREAK;

            break;

        case SEND_STARTCODE: // MAB END - STARTCODE START
            TMR0L = DELAY_FRAME;

            // Send Startcode
            DMX_TXREG = 0x00;

            DMXCurrentChannel = 0;
            DMX512EnableBlinker();
            DMXState = SEND_DATA;
            break;

        case SEND_MAB: // BREAK END - MAB START
            TMR0L = DELAY_MAB;

            DMX_BREAK_TRIS = INPUT_PIN;
            DMXState = SEND_STARTCODE;
            break;

        case SEND_BREAK: // BREAK START
            TMR0L = DELAY_BREAK;

            // Set BREAK Pin as output and hold low during BREAK
            DMX_BREAK = 0;
            DMX_BREAK_TRIS = OUTPUT_PIN;

            // Do not memcopy the DMX packet while SPIDMA reception in progress
            if (SPIDMA_SS)
            {
                // Check if any DMX data there to transmit
                if (SPIRXPacketPointer->DMXDataLength == 0)
                    break;

                // Copy SPIDMA buffer to DMX buffer
                memcpy(DMXTXPacketPointer, SPIRXPacketPointer, sizeof(SPIRXPacket));
                DMXState = SEND_MAB;
            }

            break;
    }

    // Clear Timer0 Interrupt
    INTCONbits.T0IF = 0;
}

// Timer1 interrupt for DMX LED
void DMX512LEDInterrupt(void)
{
    if (PIR1bits.TMR1IF)
    {
        // Preset Timer1
        TMR1H = DMXLED_BLINKSPEED;
        TMR1L = 0;

        if (DMXLEDCounter == 4)
            DMXLEDCounter = 0;

        if (DMXLEDCounter == 0)
            DMX_LED = 1;
        else
            DMX_LED = 0;

        DMXLEDCounter++;

        // Clear Timer1 Interrupt
        PIR1bits.TMR1IF = 0;
    }
}

// Enables DMX LED blinker
static void DMX512EnableBlinker(void)
{
    // Do not try to enable again if already enabled
    if (T1CONbits.TMR1ON)
        return;
    
    // Preset Timer1
    TMR1H = DMXLED_BLINKSPEED;
    TMR1L = 0;
    DMXLEDCounter = 0;
    DMX_LED = 1;
    PIR1bits.TMR1IF = 0;

    // Enable Timer1
    T1CONbits.TMR1ON = 1;
}

// Disables DMX LED blinker
static void DMX512DisableBlinker(void)
{
    // Do not try to disable again if already disabled
    if (!T1CONbits.TMR1ON)
        return;

    // Disable Timer1
    T1CONbits.TMR1ON = 0;

    // Turn off LED
    DMX_LED = 0;
}
